/*
 * A quick hack on parts of bb to count down to the billianth second
 * By Mark Zealey <mark@itsolve.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licences as by published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Publis License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <aalib.h>
#include <unistd.h>
#include <time.h>

aa_context *context;
aa_renderparams *params;
struct font {
	int width, height;
	char *data;
} *font;

struct font *get_font(aa_font *f)
{
	struct font *t = malloc(sizeof(struct font));
	unsigned int i, j;

	t->width = 8;
	t->height = f->height;
	t->data = calloc(1, 256 * 8 * f->height);
	for(i=0,j=0; i < 256 * f->height; i++, j += 8) {
		char c = f->data[i];
		t->data[j + 0] = c & (1 << 7);
		t->data[j + 1] = c & (1 << 6);
		t->data[j + 2] = c & (1 << 5);
		t->data[j + 3] = c & (1 << 4);
		t->data[j + 4] = c & (1 << 3);
		t->data[j + 5] = c & (1 << 2);
		t->data[j + 6] = c & (1 << 1);
		t->data[j + 7] = c & (1 << 0);
	}

	return t;
}

void fastscale(char *b1, char *b2, int x1, int x2, int y1, int y2, int width1, int width2, int color)
{
	int ddx1, ddx, spx = 0, ex;
	int ddy1, ddy, spy = 0, ey;
	int x;
	char *bb1 = b1;
	width2 -= x2;
	if (!x1 || !x2 || !y1 || !y2)
		return;
	ddx = x1 + x1;
	ddx1 = x2 + x2;
	if (ddx1 < ddx)
		spx = ddx / ddx1, ddx %= ddx1;
	ddy = y1 + y1;
	ddy1 = y2 + y2;
	if (ddy1 < ddy)
		spy = (ddy / ddy1) * width1, ddy %= ddy1;
	ey = -ddy1;
	for (; y2; y2--) {
		ex = -ddx1;
		for (x = x2; x; x--) {
			if (*b1)
				*b2 = color;
				b2++;
				b1 += spx;
				ex += ddx;
			if (ex > 0) {
				b1++;
				ex -= ddx1;
			}
		}
		b2 += width2;
		bb1 += spy;
		ey += ddy;
		if (ey > 0) {
			bb1 += width1;
			ey -= ddy1;
		}
		b1 = bb1;
	}
}

static inline void pscale(int x1, int y1, int x2, int y2, char *data, int w, int h, int color)
{
	float step;
	int xx1, xx2, yy1, yy2;
	if (x1 >= 0 && x2 < aa_imgwidth(context) && y1 >= 0 && y2 <= aa_imgheight(context))
		fastscale(data, context->imagebuffer + x1 + aa_imgwidth(context) * y1, w, x2 - x1, h, y2 - y1, w, aa_imgwidth(context), color);
	if (x2 <= 0 || x1 >= aa_imgwidth(context) || y2 <= 0 || y1 >= aa_imgheight(context))
		return;
	step = w / (float) (x2 - x1);
	if (x1 < 0) {
		xx1 = -step * x1;
		x1 = 0;
	}
	else
		xx1 = 0;
	if (x2 > aa_imgwidth(context)) {
		xx2 = step * (aa_imgwidth(context) - x1);
		x2 = aa_imgwidth(context) - 1;
	}
	else
		xx2 = w;
	step = h / (float) (y2 - y1);
	if (y1 < 0) {
		yy1 = -step * y1;
		y1 = 0;
	}
	else
		yy1 = 0;
	if (y2 > aa_imgheight(context)) {
		yy2 = step * (aa_imgheight(context) - y1);
		y2 = aa_imgheight(context) - 1;
	}
	else
		yy2 = h;
	fastscale(data + xx1 + yy1 * w, context->imagebuffer + x1 + aa_imgwidth(context) * y1, xx2 - xx1, x2 - x1, yy2 - yy1, y2 - y1, w, aa_imgwidth(context), color);

}
void print(int x, int y, float width, int height, int color, char *text)
{
	int i;
	for (i = 0; text[i]; i++)
		pscale(x + i * width, y, x + (i + 1) * width, y + height,
				font->data + font->width * (font->height *
					text[i]), font->width, font->height,
				color);
}

int main(int argc, char **argv)
{
	char mem[11];
	long p = 1;
	int i;

	aa_defparams.supported |= AA_NORMAL_MASK | AA_BOLD_MASK | AA_DIM_MASK;
	aa_parseoptions(NULL, NULL, &argc, argv);

	context = aa_autoinit(&aa_defparams);
	params = aa_getrenderparams();
	aa_render(context, params, 0, 0, 1, 1);
	font = get_font( &aa_font16 );

	while(p >= 0) {
		long t = 1000000000 - time(NULL);

		params->randomval = t / 300;

		if(p != t) {
			sprintf(mem, "%ld", t);
			memset(context->imagebuffer, 0, aa_imgwidth(context) *
					aa_imgheight(context));
			print(10, aa_scrheight(context) / 2 - 8, 20.0, 20,
					(65536 - t) / 256, mem);
			p = t;
		}
		aa_render(context, params, 0, 0, aa_scrwidth(context),
				aa_scrheight(context));
		aa_flush(context);
	}

	params->randomval = 0;

	for(i = 255; i > 0; i--) {
		memset(context->imagebuffer, 0, aa_imgwidth(context) *
				aa_imgheight(context));
		print(2, aa_scrheight(context) / 2 - 8, 20, 20, i, "HAPPY");
		print(2, aa_scrheight(context) / 2 + 8, 15, 20, i, "BILLENNIUM!");
		aa_render(context, params, 0, 0, aa_scrwidth(context),
				aa_scrheight(context));
		aa_flush(context);
		sleep((i / 255) * 4);
	}

	aa_close(context);

	return 1;
}
